/* libelf.h
 * a simple library to help with interpreting 64-bit mips elf executables
 * author: Luke Deller (luked@cse.unsw.edu.au)
 * date:   24/Sep/1999
 *
 * Updated for 32bit for AOS 2003
 */

#ifndef __LIBELF_H__
#define __LIBELF_H__

#include <stdint.h>

/* basic types */

/* file header */
struct Elf32_Header {
	unsigned char e_ident [16];
	uint16_t e_type;      /* Relocatable=1, Executable=2 (+ some more ..) */
	uint16_t e_machine;   /* Target architecture: MIPS=8 */
	uint32_t e_version;   /* Elf version (should be 1) */
	uint32_t e_entry;     /* Code entry point */
	uint32_t e_phoff;      /* Program header table */
	uint32_t e_shoff;      /* Section header table */
	uint32_t e_flags;     /* Flags */
	uint16_t e_ehsize;    /* ELF header size */
	uint16_t e_phentsize; /* Size of one program segment header */
	uint16_t e_phnum;     /* Number of program segment headers */
	uint16_t e_shentsize; /* Size of one section header */
	uint16_t e_shnum;     /* Number of section headers */
	uint16_t e_shstrndx;  /* Section header index of the string table
				 for section header names */
};

/* program segment header */
struct Elf32_Phdr {
	uint32_t p_type;      /* Segment type: Loadable segment = 1 */
	uint32_t p_offset;     /* Offset of segment in file */
	uint32_t p_vaddr;     /* Reqd virtual address of segment when loading */
	uint32_t p_paddr;     /* Reqd physical address of segment (ignore) */
	uint32_t p_filesz;   /* How many bytes this segment occupies in file */
	uint32_t p_memsz;    /* How many bytes this segment should occupy in
                           * memory (when loading, expand the segment by
                           * concatenating enough zero bytes to it) */
	uint32_t p_flags;     /* Flags: logical "or" of PF_ constants below */
	uint32_t p_align;    /* Reqd alignment of segment in memory */
};

/* constants for Elf32_Phdr.p_flags */
#define PF_X		1 /* readable segment */
#define PF_W		2 /* writeable segment */
#define PF_R		4 /* executable segment */

/* constants for indexing into Elf64_Header_t.e_ident */
#define EI_MAG0		0
#define EI_MAG1		1
#define EI_MAG2		2
#define EI_MAG3		3
#define EI_CLASS	4
#define EI_DATA		5
#define EI_VERSION	6

#define ELFMAG0         '\177'
#define ELFMAG1         'E'
#define ELFMAG2         'L'
#define ELFMAG3         'F'

#define ELFCLASS32      1

/* functions provided by elf.c */

int elf_checkFile (void *elfFile);
/* Checks that elfFile points to a valid elf file.
 * Returns 0 if the elf file is valid, < 0 if invalid.
 */

struct Elf32_Phdr * elf_getProgramSegmentTable (void *elfFile);
/* Returns a pointer to the program segment table, which is an array of
 * ELF64_Phdr_t structs.  The size of the array can be found by calling
 * getNumProgramSegments.
 */

unsigned elf_getNumProgramSegments (void *elfFile);
/* Returns the number of program segments in this elf file.
 */

void elf_showDetails (void *elfFile, char *name);
/* show the details of an elf file for debugging */


#endif

